"""Provide hypha-rpc to connecting to Hypha server."""

import shortuuid

from .rpc import RPC
from .utils import ObjectProxy
from .sync import connect_to_server as connect_to_server_sync
from .sync import get_remote_service as get_remote_service_sync
from .sync import get_rtc_service as get_rtc_service_sync
from .sync import login as login_sync
from .sync import register_rtc_service as register_rtc_service_sync
from .webrtc_client import get_rtc_service, register_rtc_service
from .websocket_client import (
    connect_to_server,
    get_remote_service,
    login,
    setup_local_client,
)


class API(ObjectProxy):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._registry = {}
        self._export_handler = self._default_export_handler

    def _default_export_handler(self, obj, config=None):
        config = config or {}
        name = config.get("name", shortuuid.uuid())
        self._registry[name] = obj

    def set_export_handler(self, handler):
        self._export_handler = handler

    def export(self, obj, config=None):
        return self._export_handler(obj, config=config)

    def get_registry(self):
        return self._registry


# An placeholder object for the API
api = API()

__all__ = [
    "api",
    "RPC",
    "login",
    "connect_to_server",
    "get_remote_service",
    "login_sync",
    "connect_to_server_sync",
    "get_remote_service_sync",
    "get_rtc_service",
    "register_rtc_service",
    "register_rtc_service_sync",
    "get_rtc_service_sync",
    "setup_local_client",
]



"""Tests for the utils module."""

import os
from functools import partial
from hypha_rpc.utils import callable_sig, callable_doc, parse_service_url
import asyncio
import pytest
import random
import httpx
from hypha_rpc import connect_to_server
from hypha_rpc.utils.serve import register_asgi_service, create_openai_chat_server
from fastapi import FastAPI
from openai import AsyncOpenAI
from . import WS_SERVER_URL

from hypha_rpc.utils.launch import launch_external_services

import pytest


def test_parse_service_url():
    # Test case 1: Basic service URL
    assert parse_service_url("https://hypha.aicell.io/public/services/hypha-login") == (
        "https://hypha.aicell.io",
        "public",
        "*",
        "hypha-login",
        "*",
    )

    # Test case 2: Service URL with client_id
    assert parse_service_url(
        "https://hypha.aicell.io/public/services/client:hypha-login"
    ) == ("https://hypha.aicell.io", "public", "client", "hypha-login", "*")

    # Test case 3: Service URL with app_id
    assert parse_service_url(
        "https://hypha.aicell.io/public/services/hypha-login@app"
    ) == ("https://hypha.aicell.io", "public", "*", "hypha-login", "app")

    # Test case 4: Service URL with both client_id and app_id
    assert parse_service_url(
        "https://hypha.aicell.io/public/services/client:hypha-login@app"
    ) == ("https://hypha.aicell.io", "public", "client", "hypha-login", "app")

    # Test case 5: Service URL with trailing slash
    assert parse_service_url(
        "https://hypha.aicell.io/public/services/hypha-login/"
    ) == ("https://hypha.aicell.io", "public", "*", "hypha-login", "*")

    # Test case 6: Invalid service URL (should raise ValueError)
    with pytest.raises(ValueError):
        parse_service_url("https://hypha.aicell.io/public/hypha-login")


def test_callable_sig():
    """Test callable_sig."""

    # Function
    def func(a, b, context=None):
        return a + b

    assert callable_sig(func) == "func(a, b, context=None)"
    assert callable_sig(func, skip_context=True) == "func(a, b)"

    # Lambda function
    def lambda_func(a, b, context=None):
        return a + b

    assert callable_sig(lambda_func) == "lambda_func(a, b, context=None)"
    assert callable_sig(lambda_func, skip_context=True) == "lambda_func(a, b)"

    # Class with a __call__ method
    class CallableClass:
        def __call__(self, a, b, context=None):
            return a + b

    assert callable_sig(CallableClass) == "CallableClass(self, a, b, context=None)"
    assert callable_sig(CallableClass, skip_context=True) == "CallableClass(self, a, b)"

    # Instance of a class with a __call__ method
    callable_instance = CallableClass()
    assert callable_sig(callable_instance) == "CallableClass(a, b, context=None)"
    assert callable_sig(callable_instance, skip_context=True) == "CallableClass(a, b)"

    # Built-in function
    assert callable_sig(print) in [
        "print(*args, **kwargs)",
        "print(*args, sep=' ', end='\\n', file=None, flush=False)",
    ]
    assert callable_sig(print, skip_context=True) in [
        "print(*args, **kwargs)",
        "print(*args, sep=' ', end='\\n', file=None, flush=False)",
    ]

    # Partial function
    partial_func = partial(func, b=3)
    assert callable_sig(partial_func) == "func(a, context=None)"
    assert callable_sig(partial_func, skip_context=True) == "func(a)"


def test_callable_doc():
    """Test callable_doc."""

    # Function with docstring
    def func_with_doc(a, b):
        """This is a function with a docstring."""
        return a + b

    assert callable_doc(func_with_doc) == "This is a function with a docstring."

    # Function without docstring
    def func_without_doc(a, b):
        return a + b

    assert callable_doc(func_without_doc) is None

    # Partial function with docstring
    def partial_func_with_doc(a, b=3):
        """This is a partial function with a docstring"""
        return a + b

    partial_func = partial(partial_func_with_doc, b=3)
    assert callable_doc(partial_func) == "This is a partial function with a docstring"


@pytest.mark.asyncio
async def test_serve_fastapi_app(websocket_server):
    """Test the serve command utility function."""
    # Create a FastAPI app instance
    app = FastAPI()

    @app.get("/")
    async def root():
        return {"message": "Hello, World!"}

    @app.get("/api/v1/test")
    async def test():
        return {"message": "Hello, it works!"}

    # Register the FastAPI app as a service
    service_id = "test-service"

    server = await connect_to_server({"server_url": WS_SERVER_URL})
    workspace = server.config["workspace"]
    token = await server.generate_token()
    await register_asgi_service(server, service_id, app)

    # Test the service using httpx
    async with httpx.AsyncClient() as client:
        response = await client.get(f"{WS_SERVER_URL}/{workspace}/apps/{service_id}/")
        assert response.status_code == 200
        assert response.json() == {"message": "Hello, World!"}

        response = await client.get(
            f"{WS_SERVER_URL}/{workspace}/apps/{service_id}/api/v1/test"
        )
        assert response.status_code == 200
        assert response.json() == {"message": "Hello, it works!"}

    try:
        # Ensure the server is no longer running
        response = await client.get(f"{WS_SERVER_URL}/{workspace}/apps/{service_id}/")
        assert response.status_code == 404
    except RuntimeError as exp:
        assert "the client has been closed" in str(exp)


@pytest.mark.asyncio
async def test_openai_server_proxy(websocket_server):
    """Test the OpenAI server proxy."""

    async def text_generator(request: dict):
        max_tokens = request.get("max_tokens", 50)
        words = [
            "hello",
            "world",
            "foo",
            "bar",
            "chatbot",
            "test",
            "api",
            "response",
            "markdown",
        ]
        # Simulate streaming random markdown text
        for _ in range(5):  # Stream 5 chunks of random text
            delta_text = " ".join(random.choices(words, k=max_tokens // 5))
            yield delta_text
            await asyncio.sleep(0.1)  # Simulate delay between chunks

    model_registry = {"test-chat-model": text_generator}
    app = create_openai_chat_server(model_registry)

    server = await connect_to_server({"server_url": WS_SERVER_URL})
    workspace = server.config["workspace"]
    token = await server.generate_token()

    service_id = "openai-server"

    await register_asgi_service(server, service_id, app)

    # Let's test it using httpx
    async with httpx.AsyncClient() as client:
        response = await client.get(
            f"{WS_SERVER_URL}/{workspace}/apps/{service_id}/v1/models"
        )
        assert response.status_code == 200
        data = response.json()
        assert data["object"] == "list"
        assert data["data"][0]["id"] == "test-chat-model"

        response = await client.post(
            f"{WS_SERVER_URL}/{workspace}/apps/{service_id}/v1/chat/completions",
            json={
                "model": "test-chat-model",
                "messages": [{"role": "user", "content": "Tell me a story."}],
                "temperature": 0.8,
                "max_tokens": 50,
                "stream": False,
            },
        )
        assert response.status_code == 200
        assert response.headers["content-type"].startswith("application/json")
        data = response.json()
        assert data["model"] == "test-chat-model"
        assert data["choices"][0]["message"]["role"] == "assistant"

    client = AsyncOpenAI(
        base_url=f"{WS_SERVER_URL}/{workspace}/apps/{service_id}/v1", api_key=token
    )
    response = await client.chat.completions.create(
        model="test-chat-model",
        messages=[{"role": "user", "content": "Tell me a story."}],
        temperature=0.8,
        max_tokens=50,
        stream=True,
    )

    async for chunk in response:
        print("Chunk:", chunk.choices[0].delta.content)
        assert chunk.choices[0].delta.role == "assistant"


@pytest.mark.asyncio
async def test_launch_external_services(websocket_server):
    """Test the launch command utility fuction."""
    server = await connect_to_server(
        {
            "name": "my third app",
            "server_url": WS_SERVER_URL,
        }
    )
    current_dir = os.path.dirname(os.path.abspath(__file__))
    proc = await launch_external_services(
        server,
        "python "
        + current_dir
        + "/example_service_script.py --server-url={server_url} --service-id=external-test-service --workspace={workspace} --token={token}",
        name="example_service_script",
        check_services=["external-test-service"],
    )
    external_service = await server.get_service(f"external-test-service")
    assert external_service.id.endswith(":external-test-service")
    assert await external_service.test(1) == 100
    await proc.kill()
    await asyncio.sleep(0.1)
    try:
        await server.get_service("external-test-service")
    except Exception as e:
        assert "not found" in str(e)
    proc = await launch_external_services(
        server,
        "python -m http.server 9391",
        name="example_service_script",
        check_url="http://127.0.0.1:9391",
    )
    assert await proc.ready()
    await proc.kill()



import os
import httpx
from simpervisor import SupervisedProcess


async def _http_ready_func(url, p, logger=None):
    """Check if the http service is ready."""
    async with httpx.AsyncClient(timeout=20.0) as client:
        try:
            resp = await client.get(url)
            # We only care if we get back *any* response, not just 200
            # If there's an error response, that can be shown directly to the user
            if logger:
                logger.debug(f"Got code {resp.status} back from {url}")
            return True
        except httpx.RequestError as exc:
            if logger:
                logger.debug(f"Connection to {url} failed: {exc}")
            return False


async def launch_external_services(
    server: any,
    command: str,
    name=None,
    env=None,
    check_services=None,
    check_url=None,
    timeout=5,
    logger=None,
):
    """
    Launch external services asynchronously and monitors their status (requires simpervisor).

    Args:
        server: The server instance for which the services are to be launched.
        command (str): The command to be executed to start the service. Any placeholders such as {server_url},
                       {workspace}, and {token} in the command will be replaced with actual values.
        name (str, optional): The name of the service. If not provided, the first argument of the command is used as the name.
        env (dict, optional): A dictionary of environment variables to be set for the service.
        check_services (list, optional): A list of service IDs to be checked for readiness. The service is considered ready
                                          if all services in the list are available.
        check_url (str, optional): A URL to be checked for readiness. The service is considered ready if the URL is accessible.
        timeout (int, optional): The maximum number of seconds to wait for the service to be ready. Defaults to 5.
        logger (logging.Logger, optional): A logger instance to be used for logging messages. If not provided, no messages
                                           are logged.

    Raises:
        Exception: If the service fails to start or does not become ready within the specified timeout.

    Returns:
        proc (SupervisedProcess): The process object for the service. You can call proc.kill() to kill the process.
    """

    token = await server.generate_token()
    # format the cmd so we fill in the {server_url} placeholder
    command = command.format(
        server_url=server.config.local_base_url,
        workspace=server.config["workspace"],
        token=token,
    )
    command = [c.strip() for c in command.split() if c.strip()]
    assert len(command) > 0, f"Invalid command: {command}"
    name = name or command[0]
    # split command into list, strip off spaces and filter out empty strings
    server_env = os.environ.copy()
    server_env.update(env or {})

    async def ready_function(p):
        if check_services:
            for service_id in check_services:
                try:
                    await server.get_service(
                        server.config["workspace"] + "/" + service_id
                    )
                except Exception:
                    return False
            return True
        if check_url:
            return await _http_ready_func(check_url, p, logger=logger)
        return True

    proc = SupervisedProcess(
        name,
        *command,
        env=server_env,
        always_restart=False,
        ready_func=ready_function,
        ready_timeout=timeout,
        log=logger,
    )

    try:
        await proc.start()

        is_ready = await proc.ready()

        if not is_ready:
            await proc.kill()
            raise Exception(f"External services ({name}) failed to start")
    except:
        if logger:
            logger.exception(f"External services ({name}) failed to start")
        raise
    return proc



"""Test the schema generation."""

import pytest
from hypha_rpc import (
    connect_to_server,
)
from pydantic import BaseModel, Field
from hypha_rpc.utils.schema import (
    schema_service,
    schema_function,
    schema_method,
    Field as NativeField,
)
from typing import Optional, Union

from . import WS_SERVER_URL


@schema_function(schema_type="native:strict")
def register_user_native(
    user_info: dict = NativeField(
        ..., description="Information of the user to register"
    ),
    receive_newsletter: bool = Field(
        False, description="Whether the user wants to receive newsletters"
    ),
) -> str:
    """Register a new user."""
    assert isinstance(user_info, dict)
    assert isinstance(receive_newsletter, bool)
    return f"User {user_info['name']} registered{' with newsletter subscription' if receive_newsletter else ''}"


@pytest.mark.asyncio
async def test_schema_function_native():
    # If we pass a dictionary, it should do model_validate with the pydantic type
    assert (
        register_user_native({"name": "Alice", "email": "alice@example.com"})
        == "User Alice registered"
    )
    assert (
        register_user_native(dict(name="Alice", email="alice@example.com"))
        == "User Alice registered"
    )
    assert (
        register_user_native(
            dict(name="Alice", email="alice@example.com"), receive_newsletter=True
        )
        == "User Alice registered with newsletter subscription"
    )


class OrderManagerNative:
    @schema_method(schema_type="native:auto")
    def place_order(
        self,
        product_id: str,
        quantity: Optional[int] = Field(
            1, description="Quantity of the product to order"
        ),
    ) -> dict:
        """Place an order for a product."""
        assert isinstance(quantity, int)
        return dict(product_id=product_id, quantity=quantity)


@pytest.mark.asyncio
async def test_schema_function_on_class_method_native():
    """Test using @schema_function on a class method."""
    manager = OrderManagerNative()

    place_order_with_schema = manager.place_order

    # If we don't pass any parameter
    # Instead of the Field which was set to the default value, it should use the default value of the Field
    assert place_order_with_schema("12345") == dict(product_id="12345", quantity=1)


class UserInfo(BaseModel):
    """User information."""

    name: str = Field(..., description="Name of the user")
    email: str = Field(..., description="Email of the user")


class UserInfoDetailed(BaseModel):
    """User information."""

    name: str = Field(..., description="Name of the user")
    email: str = Field(..., description="Email of the user")
    age: int = Field(..., description="Age of the user")
    address: str = Field(..., description="Address of the user")


@schema_function(schema_type="pydantic:strict")
def register_user(
    user_info: Union[UserInfo, UserInfoDetailed] = Field(
        ..., description="Information of the user to register"
    ),
    receive_newsletter: bool = NativeField(
        False, description="Whether the user wants to receive newsletters"
    ),
) -> str:
    """Register a new user."""
    assert isinstance(user_info, (UserInfo, UserInfoDetailed))
    assert isinstance(receive_newsletter, bool)
    if isinstance(user_info, UserInfoDetailed):
        return f"User {user_info.name} registered with detailed information{' with newsletter subscription' if receive_newsletter else ''}"
    return f"User {user_info.name} registered{' with newsletter subscription' if receive_newsletter else ''}"


@pytest.mark.asyncio
async def test_schema_function():
    # If we pass a dictionary, it should do model_validate with the pydantic type
    assert (
        register_user({"name": "Alice", "email": "alice@example.com"})
        == "User Alice registered"
    )

    assert (
        register_user(
            {
                "name": "Alice",
                "email": "alice@example.com",
                "age": 20,
                "address": "1234 Main St",
            }
        )
        == "User Alice registered with detailed information"
    )
    assert (
        register_user(UserInfo(name="Alice", email="alice@example.com"))
        == "User Alice registered"
    )
    assert (
        register_user(
            UserInfo(name="Alice", email="alice@example.com"), receive_newsletter=True
        )
        == "User Alice registered with newsletter subscription"
    )


class OrderManager:
    @schema_function(skip_self=True, schema_type="pydantic:auto")
    def place_order(
        self,
        product_id: str,
        quantity: Optional[Union[int, str]] = Field(
            1, description="Quantity of the product to order"
        ),
    ) -> dict:
        """Place an order for a product."""
        assert isinstance(quantity, (int, str))
        return dict(product_id=product_id, quantity=quantity)


@pytest.mark.asyncio
async def test_schema_function_on_class_method():
    """Test using @schema_function on a class method."""
    manager = OrderManager()

    place_order_with_schema = manager.place_order
    assert place_order_with_schema.__schema__ == {
        "name": "place_order",
        "description": "Place an order for a product.",
        "parameters": {
            "properties": {
                "product_id": {"description": "product_id", "type": "string"},
                "quantity": {
                    "anyOf": [
                        {"type": "integer"},
                        {"type": "string"},
                        {"type": "null"},
                    ],
                    "default": 1,
                    "description": "Quantity of the product to order",
                },
            },
            "required": ["product_id"],
            "type": "object",
        },
    }

    # If we don't pass any parameter
    # Instead of the Field which was set to the default value, it should use the default value of the Field
    assert place_order_with_schema("12345") == dict(product_id="12345", quantity=1)


def place_order(
    product_id: str = Field(..., description="ID of the product to order"),
    quantity: int = NativeField(1, description="Quantity of the product to order"),
) -> dict:
    """Place an order for a product."""
    assert isinstance(quantity, int)
    assert isinstance(product_id, str)
    return dict(product_id=product_id, quantity=quantity)


def place_order_native(
    product_id: str = NativeField(..., description="ID of the product to order"),
    quantity: int = Field(1, description="Quantity of the product to order"),
) -> dict:
    """Place an order for a product."""
    assert isinstance(quantity, int)
    assert isinstance(product_id, str)
    return dict(product_id=product_id, quantity=quantity)


@pytest.mark.asyncio
async def test_schema_function(websocket_server):
    """Test extract schema from functions."""
    place_order_with_schema = schema_function(place_order)

    expected = "User John registered"
    assert register_user({"name": "John", "email": "john@example.com"}) == expected
    assert register_user(UserInfo(name="John", email="john@example.com")) == expected

    assert place_order_with_schema(product_id="12345", quantity=2) == dict(
        product_id="12345", quantity=2
    )

    # test workspace is an exception, so it can pass directly
    ws = await connect_to_server({"name": "my app", "server_url": WS_SERVER_URL})

    await ws.register_service(
        {
            "name": "Order Service",
            "id": "order-service",
            "description": "Service for placing orders",
            "config": {
                "visibility": "protected",
                "run_in_executor": True,
            },
            "register_user": register_user,
            "place_order": place_order_with_schema,
        }
    )
    svc = await ws.get_service("order-service")

    assert svc.register_user.__schema__
    assert svc.place_order.__schema__
    # check some fields for the json schema
    assert svc.register_user.__schema__["name"] == "register_user"
    assert svc.register_user.__schema__ == {
        "name": "register_user",
        "description": "Register a new user.",
        "parameters": {
            "$defs": {
                "UserInfo": {
                    "description": "User information.",
                    "properties": {
                        "name": {"description": "Name of the user", "type": "string"},
                        "email": {"description": "Email of the user", "type": "string"},
                    },
                    "required": ["name", "email"],
                    "type": "object",
                },
                "UserInfoDetailed": {
                    "description": "User information.",
                    "properties": {
                        "name": {"description": "Name of the user", "type": "string"},
                        "email": {"description": "Email of the user", "type": "string"},
                        "age": {"description": "Age of the user", "type": "integer"},
                        "address": {
                            "description": "Address of the user",
                            "type": "string",
                        },
                    },
                    "required": ["name", "email", "age", "address"],
                    "type": "object",
                },
            },
            "properties": {
                "user_info": {
                    "anyOf": [
                        {"$ref": "#/$defs/UserInfo"},
                        {"$ref": "#/$defs/UserInfoDetailed"},
                    ],
                    "description": "Information of the user to register",
                },
                "receive_newsletter": {
                    "default": False,
                    "description": "Whether the user wants to receive newsletters",
                    "type": "boolean",
                },
            },
            "required": ["user_info"],
            "type": "object",
        },
    }
    assert (
        await svc.register_user({"name": "John", "email": "john@example.com"})
        == expected
    )
    assert await svc.place_order(product_id="12345", quantity=2) == dict(
        product_id="12345", quantity=2
    )


@pytest.mark.asyncio
async def test_schema_service(websocket_server):
    """Test creating schema service."""
    ws = await connect_to_server({"name": "my app", "server_url": WS_SERVER_URL})

    def place_order_with_context(
        product_id: str = Field(..., description="ID of the product to order"),
        quantity: int = Field(1, description="Quantity of the product to order"),
        context=None,
    ) -> dict:
        """Place an order for a product."""
        return dict(product_id=product_id, quantity=quantity)

    svc = await ws.register_service(
        schema_service(
            schema_type="pydantic",
            id="test",
            name="Test Service",
            description="Services for testing.",
            config={
                "require_context": True,
                "visibility": "public",
            },
            place_order=place_order_with_context,
            inner_service={
                "place_order": place_order_with_context,
            },
        )
    )
    assert svc.service_schema.place_order["type"] == "function"
    assert svc.service_schema.place_order["function"] == {
        "name": "place_order",
        "description": "Place an order for a product.",
        "parameters": {
            "properties": {
                "product_id": {
                    "description": "ID of the product to order",
                    "type": "string",
                },
                "quantity": {
                    "default": 1,
                    "description": "Quantity of the product to order",
                    "type": "integer",
                },
            },
            "required": ["product_id"],
            "type": "object",
        },
    }

    test_service = await ws.get_service(svc.id)

    assert test_service.place_order.__schema__


@pytest.mark.asyncio
async def test_schema_service_modes(websocket_server):
    """Test creating schema service with different modes."""
    ws = await connect_to_server({"name": "my app", "server_url": WS_SERVER_URL})

    # Pydantic mode
    svc_pydantic = await ws.register_service(
        schema_service(
            schema_type="pydantic",
            id="test_pydantic",
            name="Test Service Pydantic",
            description="Services for testing pydantic mode.",
            config={
                "require_context": True,
                "visibility": "public",
            },
            place_order=place_order,
            inner_service={
                "place_order": place_order,
            },
        )
    )
    assert svc_pydantic.service_schema.place_order["type"] == "function"
    assert svc_pydantic.service_schema.place_order["function"]["name"] == "place_order"
    assert "parameters" in svc_pydantic.service_schema.place_order["function"]

    # native mode
    svc_native = await ws.register_service(
        schema_service(
            schema_type="native",
            id="test_native",
            name="Test Service native",
            description="Services for testing native mode.",
            config={
                "require_context": True,
                "visibility": "public",
            },
            place_order=place_order_native,
            inner_service={
                "place_order": place_order_native,
            },
        )
    )
    assert svc_native.service_schema.place_order["type"] == "function"
    assert svc_native.service_schema.place_order["function"]["name"] == "place_order"
    assert "parameters" in svc_native.service_schema.place_order["function"]

    test_service_pydantic = await ws.get_service(svc_pydantic.id)
    assert test_service_pydantic.place_order.__schema__

    test_service_native = await ws.get_service(svc_native.id)
    assert test_service_native.place_order.__schema__ == {
        "name": "place_order",
        "description": "Place an order for a product.",
        "parameters": {
            "type": "object",
            "properties": {
                "product_id": {
                    "type": "str",
                    "description": "ID of the product to order",
                },
                "quantity": {
                    "type": "int",
                    "default": 1,
                    "description": "Quantity of the product to order",
                },
            },
            "required": ["product_id"],
        },
    }
    assert "name" in test_service_native.place_order.__schema__
    assert "description" in test_service_native.place_order.__schema__
    assert "parameters" in test_service_native.place_order.__schema__



import argparse
import importlib
import asyncio
from hypha_rpc import connect_to_server, login
import time
import random
from typing import List, Literal, Union, Optional, Callable, Dict
from fastapi import FastAPI, HTTPException
from fastapi.middleware.cors import CORSMiddleware
from pydantic import BaseModel, Field
from sse_starlette.sse import EventSourceResponse

# Define the model registry type for better type hints
ModelRegistry = Dict[str, callable]


# Base API Models
class ModelCard(BaseModel):
    id: str
    object: str = "model"
    created: int = Field(default_factory=lambda: int(time.time()))
    owned_by: str = "owner"
    root: Optional[str] = None
    parent: Optional[str] = None
    permission: Optional[list] = None


class ModelList(BaseModel):
    object: str = "list"
    data: List[ModelCard] = []


class ChatMessageInput(BaseModel):
    role: Literal["user", "assistant", "system"]
    content: Union[str, List[str]]
    name: Optional[str] = None


class ChatMessageResponse(BaseModel):
    role: Literal["assistant"]
    content: str = None
    name: Optional[str] = None


class DeltaMessage(BaseModel):
    role: Optional[Literal["user", "assistant", "system"]] = None
    content: Optional[str] = None


class ChatCompletionRequest(BaseModel):
    model: str
    messages: List[ChatMessageInput]
    temperature: Optional[float] = 0.8
    top_p: Optional[float] = 0.8
    max_tokens: Optional[int] = 50
    stream: Optional[bool] = False


class ChatCompletionResponseChoice(BaseModel):
    index: int
    message: ChatMessageResponse


class ChatCompletionResponseStreamChoice(BaseModel):
    index: int
    delta: DeltaMessage


class ChatCompletionResponse(BaseModel):
    model: str
    object: Literal["chat.completion", "chat.completion.chunk"]
    choices: List[
        Union[ChatCompletionResponseChoice, ChatCompletionResponseStreamChoice]
    ]
    created: Optional[int] = Field(default_factory=lambda: int(time.time()))


# Function to create FastAPI app based on the model registry
def create_openai_chat_server(model_registry: ModelRegistry) -> FastAPI:
    app = FastAPI()

    app.add_middleware(
        CORSMiddleware,
        allow_origins=["*"],
        allow_credentials=True,
        allow_methods=["*"],
        allow_headers=["*"],
    )

    @app.get("/v1/models", response_model=ModelList)
    async def list_models():
        model_cards = [ModelCard(id=model_name) for model_name in model_registry.keys()]
        return ModelList(data=model_cards)

    @app.post("/v1/chat/completions", response_model=ChatCompletionResponse)
    async def create_chat_completion(request: ChatCompletionRequest):
        if len(request.messages) < 1 or request.messages[-1].role == "assistant":
            raise HTTPException(status_code=400, detail="Invalid request")

        model_id = request.model
        if model_id not in model_registry:
            raise HTTPException(status_code=404, detail="Model not found")

        # Get the model's text generator function
        text_generator = model_registry[model_id]

        # Streaming mode
        if request.stream:
            generate = text_generator(request.dict())
            return EventSourceResponse(
                stream_chunks(generate), media_type="text/event-stream"
            )

        # Non-streaming mode
        response_text = ""
        async for chunk in text_generator(request.dict()):
            response_text += chunk
            if len(response_text) >= request.max_tokens:
                break
        response_text = response_text[: request.max_tokens]

        message = ChatMessageResponse(role="assistant", content=response_text)
        choice_data = ChatCompletionResponseChoice(index=0, message=message)
        return ChatCompletionResponse(
            model=request.model, choices=[choice_data], object="chat.completion"
        )

    return app


# Streaming helper to package text chunks in the response structure
async def stream_chunks(generator: Callable) -> dict:
    async for chunk in generator:
        if isinstance(chunk, str):
            delta = DeltaMessage(content=chunk, role="assistant")
            choice_data = ChatCompletionResponseStreamChoice(index=0, delta=delta)
            chunk_response = ChatCompletionResponse(
                model="test-chat-model",
                choices=[choice_data],
                object="chat.completion.chunk",
            )
            yield chunk_response.model_dump_json(exclude_unset=True)
        elif isinstance(chunk, dict):
            chunk_response = ChatCompletionResponse.model_validate(chunk)
            yield chunk_response.model_dump_json(exclude_unset=True)
        else:
            raise ValueError("Invalid chunk type")


async def serve_app(
    app: FastAPI,
    server_url: str,
    service_id: str,
    workspace: str = None,
    token: str = None,
    disable_ssl: bool = False,
    service_name: str = None,
):
    # Connection options
    connection_options = {
        "server_url": server_url,
        "workspace": workspace,
        "token": token,
    }
    if disable_ssl:
        connection_options["ssl"] = False

    # Connect to the Hypha server
    server = await connect_to_server(connection_options)
    svc_info = await register_asgi_service(server, service_id, app)
    print(
        f"Access your app at: {server_url}/{server.config.workspace}/apps/{svc_info['id'].split(':')[1]}"
    )
    # Keep the server running
    await server.serve()


async def register_asgi_service(server, service_id, app, **kwargs):
    async def serve_fastapi(args, context=None):
        await app(args["scope"], args["receive"], args["send"])

    svc = {
        "id": service_id,
        "name": service_id,
        "type": "asgi",
        "serve": serve_fastapi,
        "config": {"visibility": "public"},
    }
    svc.update(kwargs)
    svc_info = await server.register_service(svc)
    return svc_info


async def main(args):
    if args.login:
        if args.token:
            raise ValueError("Cannot use --token when --login is enabled.")
        login_options = {"server_url": args.server_url}
        if args.disable_ssl:
            login_options["ssl"] = False
        # Perform login to get the token
        token = await login(login_options)
    else:
        if not args.token:
            raise ValueError("Either --token or --login must be provided.")
        token = args.token

    # Import the app dynamically
    module_name, app_name = args.app.split(":")
    module = importlib.import_module(module_name)
    app = getattr(module, app_name)

    if not isinstance(app, FastAPI):
        raise TypeError("The specified app is not a FastAPI instance")

    # Start serving the app asynchronously
    await serve_app(
        app,
        args.server_url,
        args.id,
        args.workspace,
        token,
        args.disable_ssl,
        args.name,
    )


def main_entry():
    parser = argparse.ArgumentParser(description="Serve FastAPI app using Hypha")
    parser.add_argument(
        "app", type=str, help="The FastAPI app to serve (e.g., myapp:app)"
    )
    parser.add_argument("--id", type=str, required=True, help="The service ID")
    parser.add_argument(
        "--name", type=str, default=None, required=False, help="The service name"
    )
    parser.add_argument(
        "--server-url", type=str, required=True, help="The Hypha server URL"
    )
    parser.add_argument(
        "--workspace", type=str, default=None, help="The workspace to connect to"
    )
    parser.add_argument(
        "--token",
        type=str,
        default=None,
        help="The token for authentication (not needed if --login is used)",
    )
    parser.add_argument(
        "--login",
        action="store_true",
        help="Enable login to get the token (overrides --token)",
    )
    parser.add_argument(
        "--disable-ssl", action="store_true", help="Disable SSL verification"
    )

    args = parser.parse_args()

    # Run the main coroutine indefinitely
    asyncio.run(main(args))


if __name__ == "__main__":
    main_entry()
